from __future__ import annotations

import logging
from contextlib import contextmanager

from termcolor import colored


class QleverLogFormatter(logging.Formatter):
    """
    Custom formatter for logging.
    """

    def format(self, record):
        message = record.getMessage()
        if record.levelno == logging.DEBUG:
            return colored(f"{message}", "magenta")
        elif record.levelno == logging.WARNING:
            return colored(f"{message}", "yellow")
        elif record.levelno in [logging.CRITICAL, logging.ERROR]:
            return colored(f"{message}", "red")
        else:
            return message


# Custom logger.
log = logging.getLogger("qlever")
log.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(QleverLogFormatter())
log.addHandler(handler)
log_levels = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
    "NO_LOG": logging.CRITICAL + 1,
}


@contextmanager
def mute_log(level=logging.ERROR):
    """
    Temporarily mute the log, simply works as follows:

    with mute_log():
       ...
    """
    original_level = log.getEffectiveLevel()
    log.setLevel(level)
    try:
        yield
    finally:
        log.setLevel(original_level)



from __future__ import annotations

import subprocess
import time
from pathlib import Path

from termcolor import colored

from qlever.command import QleverCommand
from qlever.log import log
from qlever.util import (
    run_command,
)


class RebuildIndexCommand(QleverCommand):
    """
    Class for executing the `rebuild-index` command.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return "Rebuild the index from the current data (including updates)"

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str, list[str]]:
        return {
            "data": ["name"],
            "server": ["host_name", "port", "access_token"],
            "runtime": ["server_container"],
        }

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "--index-dir",
            type=str,
            help="Directory for the new index (default: subdirectory "
            "`rebuild.YYYY-MM-DDTHH:MM` of the current directory)",
        )
        subparser.add_argument(
            "--index-name",
            type=str,
            help="Base name of the new index (default: use the same as the "
            "current index)",
        )
        subparser.add_argument(
            "--restart-when-finished",
            action="store_true",
            default=False,
            help="When the rebuild is finished, stop the server with the old "
            "index and start it again with the new index",
        )

    def execute(self, args) -> bool:
        # Default values for arguments.
        if args.index_name is None:
            args.index_name = args.name
        if args.index_dir is None:
            timestamp = time.strftime("%Y-%m-%dT%H:%M", time.localtime())
            args.index_dir = f"rebuild.{timestamp}"
        if args.index_dir.endswith("/"):
            args.index_dir = args.index_dir[:-1]

        # Check that the index directory either does not exist or is empty.
        index_path = Path(args.index_dir)
        if index_path.exists() and any(index_path.iterdir()):
            log.error(
                f"The specified index directory '{args.index_dir}' already "
                "exists and is not empty; please specify an empty or "
                "non-existing directory"
            )
            return False

        # Split `index_dir` into path and dir name. For example, if `index_dir`
        # is `path/to/index`, then the path is `path/to` and the dir name
        # is `index`.
        #
        # NOTE: We keep this separate because we can always create a
        # subdirectory in the current directory (even when running in a
        # container), but not necessarily a directory at an arbitrary path. If
        # a path outside the current directory is desired, we move the index
        # there after it has been built.
        index_dir_path = str(Path(args.index_dir).parent)
        index_dir_name = str(Path(args.index_dir).name)
        log_file_name = f"{args.index_name}.rebuild-index-log.txt"

        # Command for rebuilding the index.
        mkdir_cmd = (
            f"mkdir -p {index_dir_name} && "
            f"> {index_dir_name}/{log_file_name} && "
            f"cp -a Qleverfile {index_dir_name}"
        )
        rebuild_index_cmd = (
            f"curl -s {args.host_name}:{args.port} "
            f"-d cmd=rebuild-index "
            f"-d index-name={index_dir_name}/{args.index_name} "
            f"-d access-token={args.access_token}"
        )
        move_index_cmd = f"mv {index_dir_name} {index_dir_path}"
        restart_server_cmd = (
            f"cd {args.index_dir} && "
            f"qlever start --kill-existing-with-same-port"
        )

        # Show the command lines.
        cmds_to_show = [mkdir_cmd, rebuild_index_cmd]
        if index_dir_path != ".":
            cmds_to_show.append(move_index_cmd)
        if args.restart_when_finished:
            cmds_to_show.append(restart_server_cmd)
        self.show("\n".join(cmds_to_show), only_show=args.show)
        if args.show:
            return True

        # Create the index directory and the log file.
        try:
            run_command(mkdir_cmd)
        except Exception as e:
            log.error(f"Creating the index directory failed: {e}")
            return False

        # Show the server log while rebuilding the index.
        #
        # NOTE: This will only work satisfactorily when no other queries are
        # being processed at the same time. It would be better if QLever
        # logged the rebuild-index output to a separate log file.
        tail_cmd = f"exec tail -n 0 -f {index_dir_name}/{log_file_name}"
        tail_proc = subprocess.Popen(tail_cmd, shell=True)

        # Run the index rebuild command (and time it).
        try:
            time_start = time.monotonic()
            try:
                run_command(rebuild_index_cmd, show_output=False)
            except Exception as e:
                log.error(f"Rebuilding the index failed: {e}")
                return False
            time_end = time.monotonic()
            duration_seconds = round(time_end - time_start)
            log.info("")
            rebuild_done_msg = f"Rebuilt index in {duration_seconds:,} seconds"
            if index_dir_path == ".":
                rebuild_done_msg += (
                    f", in the new directory '{args.index_dir}'"
                )
            log.info(rebuild_done_msg)
        finally:
            tail_proc.terminate()
            tail_proc.wait()

        # Move the new index to the specified directory, if needed.
        if index_dir_path != ".":
            try:
                log.info(f"Moving the new index to {args.index_dir}")
                run_command(move_index_cmd)
            except Exception as e:
                log.error(f"Moving the new index failed: {e}")
                return False

        # Restart the server with the new index, if requested.
        if args.restart_when_finished:
            try:
                log.info("Restarting the server with the new index ...")
                log.info("")
                log.info(colored("Command: start", attrs=["bold"]))
                log.info("")
                run_command(restart_server_cmd, show_output=True)
            except Exception as e:
                log.error(f"Restarting the server failed: {e}")
                return False

        return True



from __future__ import annotations

import re

from qlever.command import QleverCommand
from qlever.log import log
from qlever.util import run_command


class ResetUpdatesCommand(QleverCommand):
    """
    Class for executing the `reset-updates` command.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return "Reset the updates on the server"

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str, list[str]]:
        return {"server": ["host_name", "port", "access_token"]}

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "--sparql-endpoint",
            help="URL of the QLever server, default is {host_name}:{port}",
        )

    def execute(self, args) -> bool:
        reset_cmd = "curl -s"
        if args.sparql_endpoint:
            reset_cmd += f" {args.sparql_endpoint}"
        else:
            reset_cmd += f" {args.host_name}:{args.port}"
        reset_cmd += f' --data-urlencode "cmd=clear-delta-triples" --data-urlencode "access-token={args.access_token}"'
        self.show(reset_cmd, only_show=args.show)
        if args.show:
            return True

        try:
            reset_cmd += ' -w " %{http_code}"'
            result = run_command(reset_cmd, return_output=True)
            match = re.match(r"^(.*) (\d+)$", result, re.DOTALL)
            if not match:
                raise Exception(f"Unexpected output:\n{result}")
            error_message = match.group(1).strip()
            status_code = match.group(2)
            if status_code != "200":
                raise Exception(error_message)
            message = "Updates reset successfully"
            log.info(message)
            return True
        except Exception as e:
            log.error(e)
            return False



from __future__ import annotations

from os import environ
from pathlib import Path

import yaml

from qlever.command import QleverCommand
from qlever.containerize import Containerize
from qlever.log import log
from qlever.util import is_port_used, run_command


# Return a YAML string for the given dictionary. Format values with
# newlines using the "|" style.
def dict_to_yaml(dictionary: dict) -> str:
    """
    Custom representer for yaml, which uses the "|" style only for
    multiline strings.

    NOTE: We replace all `\r\n` with `\n` because otherwise the `|` style
    does not work as expected.
    """

    class MultiLineDumper(yaml.SafeDumper):
        def represent_scalar(self, tag, value, style=None):
            value = value.replace("\r\n", "\n")
            if isinstance(value, str) and "\n" in value:
                style = "|"
            return super().represent_scalar(tag, value, style)

    # Dump as yaml.
    return yaml.dump(
        dictionary,
        sort_keys=False,
        allow_unicode=True,
        Dumper=MultiLineDumper,
    )


class UiCommand(QleverCommand):
    """
    Class for launching the QLever UI web application.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return "Launch the QLever UI web application"

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str, list[str]]:
        return {
            "data": ["name"],
            "server": ["host_name", "port"],
            "ui": [
                "ui_port",
                "ui_config",
                "ui_system",
                "ui_image",
                "ui_container",
            ],
        }

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "--ui-config-file",
            default="Qleverfile-ui.yml",
            help="Name of the config file for the QLever UI "
            "(default: Qleverfile-ui.yml)",
        )
        subparser.add_argument(
            "--ui-db-file",
            help="Name of the database file for the QLever UI "
            "(default: {name}.ui-db.sqlite3)",
        )
        subparser.add_argument(
            "--no-pull-latest",
            action="store_true",
            default=False,
            help="Do not pull the latest image for the QLever UI "
            "(default: pull the latest image if image name contains '/')",
        )
        subparser.add_argument(
            "--stop",
            action="store_true",
            default=False,
            help="Stop the running container",
        )

    def execute(self, args) -> bool:
        # If QLEVER_OVERRIDE_DISABLE_UI is set, this command is disabled.
        qlever_is_running_in_container = environ.get(
            "QLEVER_IS_RUNNING_IN_CONTAINER"
        )
        if qlever_is_running_in_container:
            log.error(
                "The environment variable `QLEVER_OVERRIDE_DISABLE_UI` is set, "
                "therefore `qlever ui` is not available (it should not be called "
                "from inside a container)"
            )
            log.info("")
            if not args.show:
                log.info(
                    "For your information, showing the commands that are "
                    "executed when `qlever ui` is available:"
                )
                log.info("")

        # Construct commands and show them.
        pull_latest_image = "/" in args.ui_image and not args.no_pull_latest
        ui_config_name = args.ui_config
        ui_db_file = args.ui_db_file or f"{args.name}.ui-db.sqlite3"
        ui_db_file_from_image = "qleverui.sqlite3"
        ui_config_file = args.ui_config_file
        sparql_endpoint = f"http://{args.host_name}:{args.port}"
        ui_url = f"http://{args.host_name}:{args.ui_port}"
        pull_cmd = f"{args.ui_system} pull -q {args.ui_image}"
        get_db_cmd = (
            f"{args.ui_system} create "
            f"--name {args.ui_container} "
            f"{args.ui_image} "
            f"&& {args.ui_system} cp "
            f"{args.ui_container}:/app/db/{ui_db_file_from_image} {ui_db_file} "
            f"&& {args.ui_system} rm -f {args.ui_container}"
        )
        start_ui_cmd = (
            f"{args.ui_system} run -d "
            f"--volume $(pwd):/app/db "
            f"--env QLEVERUI_DATABASE_URL=sqlite:////app/db/{ui_db_file} "
            f"--publish {args.ui_port}:7000 "
            f"--name {args.ui_container} "
            f"{args.ui_image}"
        )
        get_config_cmd = (
            f"{args.ui_system} exec -i "
            f"{args.ui_container} "
            f'bash -c "python manage.py config {ui_config_name}"'
        )
        set_config_cmd = (
            f"{args.ui_system} exec -i "
            f"{args.ui_container} "
            f'bash -c "python manage.py config {ui_config_name} '
            f'/app/db/{ui_config_file} --hide-all-other-backends"'
        )
        commands_to_show = []
        if not args.stop:
            if pull_latest_image:
                commands_to_show.append(pull_cmd)
            if not Path(ui_db_file).exists():
                commands_to_show.append(get_db_cmd)
            commands_to_show.append(start_ui_cmd)
            if not Path(ui_config_file).exists():
                commands_to_show.append(get_config_cmd)
            else:
                commands_to_show.append(set_config_cmd)
        self.show("\n".join(commands_to_show), only_show=args.show)
        if qlever_is_running_in_container:
            return False
        if args.show:
            return True

        # Stop running containers.
        was_found_and_stopped = False
        for container_system in Containerize.supported_systems():
            was_found_and_stopped |= Containerize.stop_and_remove_container(
                container_system, args.ui_container
            )
        if was_found_and_stopped:
            log.debug(f"Stopped and removed container `{args.ui_container}`")
        else:
            log.debug(f"No container with name `{args.ui_container}` found")
        if args.stop:
            return True

        # Pull the latest image.
        if pull_latest_image:
            log.debug(f"Pulling image `{args.ui_image}` for QLever UI")
            run_command(pull_cmd)

        # Check if the UI port is already being used.
        if is_port_used(args.ui_port):
            log.warning(
                f"It looks like port {args.ui_port} for the QLever UI "
                f"is already in use. You can set another port in the "
                f" Qleverfile in the [ui] section with the UI_PORT variable."
            )

        # Get the QLever UI database from the image, unless it already exists.
        if Path(ui_db_file).exists():
            log.debug(f"Found QLever UI database `{ui_db_file}`, reusing it")
        else:
            log.debug(f"Getting QLever UI database `{ui_db_file}` from image")
            try:
                run_command(get_db_cmd)
            except Exception as e:
                log.error(
                    f"Failed to get {ui_db_file} from {args.ui_image} "
                    f"({e})"
                )
                return False

        # Start the QLever UI.
        try:
            log.debug(
                f"Starting new container with name `{args.ui_container}`"
            )
            run_command(start_ui_cmd)
        except Exception as e:
            log.error(f"Failed to start container `{args.ui_container}` ({e})")
            return False

        # Check if config file with name `ui_config_file` exists. If not, try
        # to obtain it via `get_config_cmd` and set it as default.
        if Path(ui_config_file).exists():
            log.info(f"Found config file `{ui_config_file}` and reusing it")
        else:
            try:
                log.info(
                    f"Get default config file `{ui_config_file}` from image "
                    f"`{args.ui_image}` and set endpoint to `{sparql_endpoint}`"
                )
                config_yaml = run_command(get_config_cmd, return_output=True)
                config_dict = yaml.safe_load(config_yaml)
            except Exception as e:
                log.error("")
                log.error(
                    f"An error occured while getting and parsing the "
                    f"config file ({e})"
                )
                return False
            try:
                config_dict["config"]["backend"]["isDefault"] = True
                config_dict["config"]["backend"]["baseUrl"] = sparql_endpoint
                config_dict["config"]["backend"]["sortKey"] = 1
                config_yaml = dict_to_yaml(config_dict)
                with open(ui_config_file, "w") as f:
                    f.write(config_yaml)
            except Exception as e:
                log.error("")
                log.error(
                    f"An error occured while modifying and writing the "
                    f"config file ({e})"
                )
                return False

        # Configure the QLever UI.
        try:
            run_command(set_config_cmd)
        except Exception as e:
            log.error(f"Failed to configure the QLever UI ({e})")
            return False

        # If we come this far, everything should work.
        log.info("")
        log.info(
            f"The QLever UI should now be up at {ui_url}/{ui_config_name}"
        )
        log.info("")
        log.debug(
            "If you must, you can log in as QLever UI admin with "
            'username and password "demo"'
        )
        log.info(
            f"You can modify the config file at `{ui_config_file}` "
            f"and then just run `qlever ui` again"
        )
        return True



from __future__ import annotations

import re
import socket
import subprocess
from configparser import ConfigParser, ExtendedInterpolation, RawConfigParser
from pathlib import Path

from qlever import script_name
from qlever.containerize import Containerize
from qlever.log import log


class QleverfileException(Exception):
    pass


class Qleverfile:
    """
    Class that defines all the possible parameters that can be specified in a
    Qleverfile + functions for parsing.
    """

    # Runtime parameters (for `settings` and `start` commands).
    SERVER_RUNTIME_PARAMETERS = [
        "cache-max-num-entries",
        "cache-max-size",
        "cache-max-size-single-entry",
        "cache-service-results",
        "default-query-timeout",
        "division-by-zero-is-undef",
        "enable-distributive-union",
        "enable-prefilter-on-index-scans",
        "group-by-disable-index-scan-optimizations",
        "group-by-hash-map-enabled",
        "lazy-index-scan-max-size-materialization",
        "lazy-index-scan-num-threads",
        "lazy-index-scan-queue-size",
        "lazy-result-max-cache-size",
        "query-planning-budget",
        "request-body-limit",
        "service-max-redirects",
        "service-max-value-rows",
        "sort-estimate-cancellation-factor",
        "sparql-results-json-with-time",
        "spatial-join-prefilter-max-size",
        "spatial-join-max-num-threads",
        "strip-columns",
        "syntax-test-mode",
        "throw-on-unbound-variables",
        "treat-default-graph-as-named-graph",
        "use-binsearch-transitive-path",
        "websocket-updates-enabled",
    ]

    @staticmethod
    def all_arguments():
        """
        Define all possible parameters. A value of `None` means that there is
        no default value.
        """

        # Helper function that takes a list of positional arguments and a list
        # of keyword arguments and returns a tuple of both. That way, we can
        # defined arguments below with exactly the same syntax as we would for
        # `argparse.add_argument`.
        def arg(*args, **kwargs):
            return (args, kwargs)

        all_args = {}
        data_args = all_args["data"] = {}
        index_args = all_args["index"] = {}
        server_args = all_args["server"] = {}
        runtime_args = all_args["runtime"] = {}
        ui_args = all_args["ui"] = {}

        data_args["name"] = arg(
            "--name", type=str, required=True, help="The name of the dataset"
        )
        data_args["get_data_cmd"] = arg(
            "--get-data-cmd",
            type=str,
            required=True,
            help="The command to get the data",
        )
        data_args["description"] = arg(
            "--description",
            type=str,
            required=True,
            help="A concise description of the dataset",
        )
        data_args["text_description"] = arg(
            "--text-description",
            type=str,
            default=None,
            help="A concise description of the additional text data if any",
        )
        data_args["format"] = arg(
            "--format",
            type=str,
            default="ttl",
            choices=["ttl", "nt", "nq"],
            help="The format of the data",
        )

        index_args["input_files"] = arg(
            "--input-files",
            type=str,
            required=True,
            help="A space-separated list of patterns that match "
            "all the files of the dataset",
        )
        index_args["cat_input_files"] = arg(
            "--cat-input-files",
            type=str,
            help="The command that produces the input",
        )
        index_args["multi_input_json"] = arg(
            "--multi-input-json",
            type=str,
            default=None,
            help="JSON to specify multiple input files, each with a "
            "`cmd` (command that writes the triples to stdout), "
            "`format` (format like for the `--format` option), "
            "`graph` (name of the graph, use `-` for the default graph), "
            "`parallel` (parallel parsing for large files, where all "
            "prefix declaration are at the beginning)",
        )
        index_args["parallel_parsing"] = arg(
            "--parallel-parsing",
            type=str,
            choices=["true", "false"],
            help="Use parallel parsing (recommended for large files, "
            "but it requires that all prefix declarations are at the "
            "beginning of the file)",
        )
        index_args["settings_json"] = arg(
            "--settings-json",
            type=str,
            default="{}",
            help="The `.settings.json` file for the index",
        )
        index_args["ulimit"] = arg(
            "--ulimit",
            type=int,
            default=None,
            help="Explicitly set the limit for the maximal number of open "
            "files (default: 1048576 when the total size of the input files "
            "is larger than 10 GB)",
        )
        index_args["vocabulary_type"] = arg(
            "--vocabulary-type",
            type=str,
            choices=[
                "on-disk-compressed",
                "on-disk-uncompressed",
                "in-memory-compressed",
                "in-memory-uncompressed",
                "on-disk-compressed-geo-split",
            ],
            default="on-disk-compressed",
            help="The type of the vocabulary to use for the index "
            " (default: `on-disk-compressed`)",
        )
        index_args["index_binary"] = arg(
            "--index-binary",
            type=str,
            default="IndexBuilderMain",
            help="The binary for building the index (this requires "
            "that you have compiled QLever on your machine)",
        )
        index_args["stxxl_memory"] = arg(
            "--stxxl-memory",
            type=str,
            help="The amount of memory to use for the index build "
            "(the name of the option has historical reasons)",
        )
        index_args["parser_buffer_size"] = arg(
            "--parser-buffer-size",
            type=str,
            help="Each parser thread reads the input in batches of this size, "
            "and in parallel parsing, each batch that is not the last must be "
            "large enough to contain the end of at least one statement "
            "(default: 10M)",
        )
        index_args["encode_as_id"] = arg(
            "--encode-as-id",
            type=str,
            help="Space-separated list of IRI prefixes (without angle "
            "brackets); IRIs that start with one of these prefixes, followed "
            "by a sequence of digits, do not require a vocabulary entry but "
            "are directly encoded in the ID (default: none)",
        )
        index_args["only_pso_and_pos_permutations"] = arg(
            "--only-pso-and-pos-permutations",
            action="store_true",
            default=False,
            help="Only create the PSO and POS permutations",
        )
        index_args["use_patterns"] = arg(
            "--use-patterns",
            choices=["yes", "no"],
            default="yes",
            help="Whether to precompute the so-called patterns used for fast "
            "processing of queries like SELECT ?p (COUNT(DISTINCT ?s) AS ?c) "
            "WHERE { ?s ?p [] ... } GROUP BY ?p",
        )
        index_args["add_has_word_triples"] = arg(
            "--add-has-word-triples",
            action="store_true",
            default=False,
            help="Whether to add `ql:has-word` triples for text literals "
            "(which can then be used for custom text search queries)",
        )
        index_args["text_index"] = arg(
            "--text-index",
            choices=[
                "none",
                "from_text_records",
                "from_literals",
                "from_text_records_and_literals",
            ],
            default="none",
            help="Whether to also build an index for text search"
            "and for which texts",
        )
        index_args["text_words_file"] = arg(
            "--text-words-file",
            type=str,
            default=None,
            help="File with the words for the text index (one line "
            "per word, format: `word or IRI\t0 or 1\tdoc id\t1`)",
        )
        index_args["text_docs_file"] = arg(
            "--text-docs-file",
            type=str,
            default=None,
            help="File with the documents for the text index (one line "
            "per document, format: `id\tdocument text`)",
        )

        server_args["server_binary"] = arg(
            "--server-binary",
            type=str,
            default="ServerMain",
            help="The binary for starting the server (this requires "
            "that you have compiled QLever on your machine)",
        )
        server_args["host_name"] = arg(
            "--host-name",
            type=str,
            help="The name of the host on which the server listens for "
            "requests",
        )
        server_args["port"] = arg(
            "--port",
            type=int,
            help="The port on which the server listens for requests",
        )
        server_args["access_token"] = arg(
            "--access-token",
            type=str,
            default=None,
            help="The access token for privileged operations",
        )
        server_args["memory_for_queries"] = arg(
            "--memory-for-queries",
            type=str,
            default="5G",
            help="The maximal amount of memory used for query processing"
            " (if a query needs more than what is available, the "
            "query will not be processed)",
        )
        server_args["cache_max_size"] = arg(
            "--cache-max-size",
            type=str,
            default="2G",
            help="The maximal amount of memory used for caching",
        )
        server_args["cache_max_size_single_entry"] = arg(
            "--cache-max-size-single-entry",
            type=str,
            default="1G",
            help="The maximal amount of memory used for caching a single "
            "query result",
        )
        server_args["cache_max_num_entries"] = arg(
            "--cache-max-num-entries",
            type=int,
            default=200,
            help="The maximal number of entries in the cache"
            " (the eviction policy when the cache is full is LRU)",
        )
        server_args["timeout"] = arg(
            "--timeout",
            type=str,
            default="30s",
            help="The maximal time in seconds a query is allowed to run"
            " (can be increased per query with the URL parameters "
            "`timeout` and `access_token`)",
        )
        server_args["num_threads"] = arg(
            "--num-threads",
            type=int,
            default=8,
            help="The number of threads used for query processing",
        )
        server_args["persist_updates"] = arg(
            "--persist-updates",
            action="store_true",
            default=False,
            help="Persist updates to the index (write updates to disk and "
            "read them back in when restarting the server)",
        )
        server_args["only_pso_and_pos_permutations"] = arg(
            "--only-pso-and-pos-permutations",
            action="store_true",
            default=False,
            help="Only use the PSO and POS permutations (then each "
            "triple pattern must have a fixed predicate)",
        )
        server_args["use_patterns"] = arg(
            "--use-patterns",
            choices=["yes", "no"],
            default="yes",
            help="Whether to use the patterns precomputed during the index "
            "build (see `qlever index --help` for their utility)",
        )
        server_args["use_text_index"] = arg(
            "--use-text-index",
            choices=["yes", "no"],
            default="no",
            help="Whether to use the text index (requires that one was "
            "built, see `qlever index`)",
        )
        server_args["warmup_cmd"] = arg(
            "--warmup-cmd",
            type=str,
            help="Command executed after the server has started "
            " (executed as part of `qlever start` unless "
            " `--no-warmup` is specified, or with `qlever warmup`)",
        )

        runtime_args["system"] = arg(
            "--system",
            type=str,
            choices=Containerize.supported_systems() + ["native"],
            default="docker",
            help=(
                "Whether to run commands like `index` or `start` "
                "natively or in a container, and if in a container, "
                "which system to use"
            ),
        )
        runtime_args["image"] = arg(
            "--image",
            type=str,
            default="docker.io/adfreiburg/qlever",
            help="The name of the image when running in a container",
        )
        runtime_args["index_container"] = arg(
            "--index-container",
            type=str,
            help=f"The name of the container used by `{script_name} index`",
        )
        runtime_args["server_container"] = arg(
            "--server-container",
            type=str,
            help=f"The name of the container used by `{script_name} start`",
        )

        ui_args["ui_port"] = arg(
            "--ui-port",
            type=int,
            default=8176,
            help="The port of the Qlever UI when running `qlever ui`",
        )
        ui_args["ui_config"] = arg(
            "--ui-config",
            type=str,
            default="default",
            help="The name of the backend configuration for the QLever UI"
            " (this determines AC queries and example queries)",
        )
        ui_args["ui_system"] = arg(
            "--ui-system",
            type=str,
            choices=Containerize.supported_systems(),
            default="docker",
            help="Which container system to use for `qlever ui`"
            " (unlike for `qlever index` and `qlever start`, "
            ' "native" is not yet supported here)',
        )
        ui_args["ui_image"] = arg(
            "--ui-image",
            type=str,
            default="docker.io/adfreiburg/qlever-ui",
            help="The name of the image used for `qlever ui`",
        )
        ui_args["ui_container"] = arg(
            "--ui-container",
            type=str,
            help="The name of the container used for `qlever ui`",
        )

        return all_args

    @staticmethod
    def read(qleverfile_path):
        """
        Read the given Qleverfile (the function assumes that it exists) and
        return a `ConfigParser` object with all the options and their values.

        NOTE: The keys have the same hierarchical structure as the keys in
        `all_arguments()`. The Qleverfile may contain options that are not
        defined in `all_arguments()`. They can be used as temporary variables
        to define other options, but cannot be accessed by the commands later.
        """

        # Read the Qleverfile.
        defaults = {"random": "83724324hztz", "version": "01.01.01"}
        config = ConfigParser(
            interpolation=ExtendedInterpolation(), defaults=defaults
        )
        try:
            config.read(qleverfile_path)
        except Exception as e:
            raise QleverfileException(f"Error parsing {qleverfile_path}: {e}")

        # Iterate over all sections and options and check if there are any
        # values of the form $$(...) that need to be replaced.
        for section in config.sections():
            for option in config[section]:
                value = config[section][option]
                match = re.match(r"^\$\((.*)\)$", value)
                if match:
                    try:
                        value = subprocess.check_output(
                            match.group(1),
                            shell=True,
                            text=True,
                            stderr=subprocess.STDOUT,
                        ).strip()
                    except Exception as e:
                        log.info("")
                        log.error(
                            f"Error evaluating {value} for option "
                            f"{section}.{option.upper()} in "
                            f"{qleverfile_path}:"
                        )
                        log.info("")
                        log.info(e.output if hasattr(e, "output") else e)
                        exit(1)
                    config[section][option] = value

        # Make sure that all the sections are there.
        for section in ["data", "index", "server", "runtime", "ui"]:
            if section not in config:
                config[section] = {}

        # Add default values that are based on other values.
        if "name" in config["data"]:
            name = config["data"]["name"]
            runtime = config["runtime"]
            if "server_container" not in runtime:
                runtime["server_container"] = f"{script_name}.server.{name}"
            if "index_container" not in runtime:
                runtime["index_container"] = f"{script_name}.index.{name}"
            if "ui_container" not in config["ui"]:
                config["ui"]["ui_container"] = f"qlever.ui.{name}"
            index = config["index"]
            if "text_words_file" not in index:
                index["text_words_file"] = f"{name}.wordsfile.tsv"
            if "text_docs_file" not in index:
                index["text_docs_file"] = f"{name}.docsfile.tsv"
            server = config["server"]
        if index.get("text_index", "none") != "none":
            server["use_text_index"] = "yes"
        if index.get("only_pso_and_pos_permutations", "false") == "true":
            index["use_patterns"] = "no"
        if index.get("use_patterns", None) == "no":
            server["use_patterns"] = "no"

        # Add other non-trivial default values.
        try:
            if config["server"].get("host_name") is None:
                config["server"]["host_name"] = socket.gethostname()
        except Exception:
            log.warning(
                "Could not get the hostname, using `localhost` as default"
            )
            pass

        # Return the parsed Qleverfile with the added inherited values.
        return config

    @staticmethod
    def filter(
        qleverfile_path: Path, options_included: dict[str, list[str]]
    ) -> RawConfigParser:
        """
        Given a filter criteria (key: section_header, value: list[options]),
        return a RawConfigParser object to create a new filtered Qleverfile
        with only the specified sections and options (selects all options if
        list[options] is empty). Mainly to be used by non-qlever scripts for
        the setup-config command
        """
        # Read the Qleverfile.
        config = RawConfigParser()
        config.optionxform = str  # Preserve case sensitivity of keys
        config.read(qleverfile_path)

        filtered_config = RawConfigParser()
        filtered_config.optionxform = str

        for section, desired_fields in options_included.items():
            if config.has_section(section):
                filtered_config.add_section(section)

                # If the list is empty, copy all fields
                if not desired_fields:
                    for field, value in config.items(section):
                        filtered_config.set(section, field, value)
                else:
                    for desired_field in desired_fields:
                        if config.has_option(section, desired_field):
                            value = config.get(section, desired_field)
                            filtered_config.set(section, desired_field, value)

        return filtered_config



from __future__ import annotations

import argparse
import os
import traceback
from importlib.metadata import version
from pathlib import Path

import argcomplete
from termcolor import colored

from qlever import command_objects, engine_name, script_name
from qlever.log import log, log_levels
from qlever.qleverfile import Qleverfile


# Simple exception class for configuration errors (the class need not do
# anything, we just want a distinct exception type).
class ConfigException(Exception):
    def __init__(self, message):
        stack = traceback.extract_stack()[-2]  # Caller's frame.
        self.filename = stack.filename
        self.lineno = stack.lineno
        full_message = f"{message} [in {self.filename}:{self.lineno}]"
        super().__init__(full_message)


class QleverConfig:
    """
    Class that manages all config parameters, and overwrites them with the
    settings from a Qleverfile.

    IMPORTANT: An instance of this class is created for each execution of
    the `qlever` script, in particular, each time the user triggers
    autocompletion. Therefore, pay attention that no unnecessary work is done
    before the call to `argcomplete.autocomplete(...)`. In particular, avoid
    parsing the Qleverfile before that point, it's not needed for
    autocompletion.
    """

    def add_subparser_for_command(self, subparsers, command_name,
                                  command_object, all_qleverfile_args,
                                  qleverfile_config=None):
        """
        Add subparser for the given command. Take the arguments from
        `command_object.relevant_qleverfile_arguments()` and report an error if
        one of them is not contained in `all_qleverfile_args`. Overwrite the
        default values with the values from `qleverfile_config` if specified.
        """

        arg_names = command_object.relevant_qleverfile_arguments()

        # Helper function that shows a detailed error messahe when an argument
        # from `relevant_qleverfile_arguments` is not contained in
        # `all_qleverfile_args`.
        def argument_error(prefix):
            log.info("")
            log.error(f"{prefix} in `Qleverfile.all_arguments()` for command "
                      f"`{command_name}`")
            log.info("")
            log.info(f"Value of `relevant_qleverfile_arguments` for "
                     f"command `{command_name}`:")
            log.info("")
            log.info(f"{arg_names}")
            log.info("")
            exit(1)

        # Add the subparser.
        description = command_object.description()
        subparser = subparsers.add_parser(command_name,
                                          description=description,
                                          help=description)

        # Add the arguments relevant for the command.
        for section in arg_names:
            if section not in all_qleverfile_args:
                argument_error(f"Section `{section}` not found")
            for arg_name in arg_names[section]:
                if arg_name not in all_qleverfile_args[section]:
                    argument_error(f"Argument `{arg_name}` of section "
                                   f"`{section}` not found")
                args, kwargs = all_qleverfile_args[section][arg_name]
                kwargs_copy = kwargs.copy()
                # If `qleverfile_config` is given, add info about default
                # values to the help string.
                if qleverfile_config is not None:
                    default_value = kwargs.get("default", None)
                    qleverfile_value = qleverfile_config.get(
                            section, arg_name, fallback=None)
                    if qleverfile_value is not None:
                        kwargs_copy["default"] = qleverfile_value
                        kwargs_copy["required"] = False
                        kwargs_copy["help"] += (f" [default, from Qleverfile:"
                                                f" {qleverfile_value}]")
                    else:
                        kwargs_copy["help"] += f" [default: {default_value}]"
                subparser.add_argument(*args, **kwargs_copy)

        # Additional arguments that are shared by all commands.
        command_object.additional_arguments(subparser)
        subparser.add_argument("--show", action="store_true",
                               default=False,
                               help="Only show what would be executed"
                                    ", but don't execute it")
        subparser.add_argument("--log-level",
                               choices=log_levels.keys(),
                               default="INFO",
                               help="Set the log level")

    def parse_args(self):
        # Determine whether we are in autocomplete mode or not.
        autocomplete_mode = "COMP_LINE" in os.environ

        # Check if the user has registered this script for argcomplete.
        argcomplete_check_off = os.environ.get("QLEVER_ARGCOMPLETE_CHECK_OFF")
        argcomplete_enabled = os.environ.get("QLEVER_ARGCOMPLETE_ENABLED")
        if not argcomplete_enabled and not argcomplete_check_off:
            log.info("")
            log.warn(f"To enable autocompletion, run the following command, "
                     f"and consider adding it to your `.bashrc` or `.zshrc`:"
                     f"\n\n"
                     f"eval \"$(register-python-argcomplete {script_name})\""
                     f" && export QLEVER_ARGCOMPLETE_ENABLED=1")
            log.info("")

        # Create a temporary parser only to parse the `--qleverfile` option, in
        # case it is given, and to determine whether a command was given that
        # requires a Qleverfile. This is because in the actual parser below we
        # want the values from the Qleverfile to be shown in the help strings,
        # but only if this is actually necessary.
        def add_qleverfile_option(parser):
            parser.add_argument("--qleverfile", "-q", type=str,
                                default="Qleverfile")
        qleverfile_parser = argparse.ArgumentParser(add_help=False)
        add_qleverfile_option(qleverfile_parser)
        qleverfile_parser.add_argument("command", type=str, nargs="?")
        qleverfile_args, _ = qleverfile_parser.parse_known_args()
        qleverfile_path_name = qleverfile_args.qleverfile
        # command = qleverfile_args.command
        # should_have_qleverfile = command in command_objects \
        #     and command_objects[command].should_have_qleverfile()

        # Check if the Qleverfile exists and if we are using the default name.
        # We need this again further down in the code, so remember it.
        qleverfile_path = Path(qleverfile_path_name)
        qleverfile_exists = qleverfile_path.is_file()
        qleverfile_is_default = qleverfile_path_name \
            == qleverfile_parser.get_default("qleverfile")
        # If a Qleverfile with a non-default name was specified, but it does
        # not exist, that's an error.
        if not qleverfile_exists and not qleverfile_is_default:
            raise ConfigException(f"Qleverfile with non-default name "
                                  f"`{qleverfile_path_name}` specified, "
                                  f"but it does not exist")
        # If it exists and we are not in the autocompletion mode, parse it.
        #
        # IMPORTANT: No need to parse the Qleverfile in autocompletion mode and
        # it would be unnecessarily expensive to do so.
        #
        # TODO: What if `command.should_have_qleverfile()` is `False`, should
        # we then parse the Qleverfile or not.
        if qleverfile_exists and not autocomplete_mode:
            try:
                qleverfile_config = Qleverfile.read(qleverfile_path)
            except Exception as e:
                log.info("")
                log.error(f"Error parsing Qleverfile `{qleverfile_path}`"
                          f": {e}")
                log.info("")
                exit(1)
        else:
            qleverfile_config = None

        # Now the regular parser with commands and a subparser for each
        # command. We have a dedicated class for each command. These classes
        # are defined in the modules in `qlever/commands`. In `__init__.py`
        # an object of each class is created and stored in `command_objects`.
        parser = argparse.ArgumentParser(
            description=colored(
                f"This is the {script_name} command line tool, "
                f"it's all you need to work with {engine_name}",
                attrs=["bold"],
            )
        )
        if script_name == "qlever":
            parser.add_argument(
                "--version",
                action="version",
                version=f"%(prog)s {version('qlever')}",
            )
        add_qleverfile_option(parser)
        subparsers = parser.add_subparsers(dest='command')
        subparsers.required = True
        all_args = Qleverfile.all_arguments()
        for command_name, command_object in command_objects.items():
            self.add_subparser_for_command(
                    subparsers, command_name, command_object,
                    all_args, qleverfile_config)

        # Enable autocompletion for the commands and their options.
        #
        # NOTE: All code executed before this line should be relatively cheap
        # because it is executed whenever the user triggers the autocompletion.
        argcomplete.autocomplete(parser, always_complete_options="long")

        # If called without arguments, show the help message.
        if len(os.sys.argv) == 1:
            parser.print_help()
            exit(0)

        # Parse the command line arguments.
        args = parser.parse_args()

        # If the command says that we should have a Qleverfile, but we don't,
        # issue a warning.
        if command_objects[args.command].should_have_qleverfile():
            if not qleverfile_exists:
                log.warning(f"Invoking command `{args.command}` without a "
                            "Qleverfile. You have to specify all required "
                            "arguments on the command line. This is possible, "
                            "but not recommended.")

        return args



from __future__ import annotations

import shlex
import time
import traceback

from qlever.command import QleverCommand
from qlever.log import log
from qlever.util import run_command


class QueryCommand(QleverCommand):
    """
    Class for executing the `query` command.
    """

    def __init__(self):
        self.predefined_queries = {
            "all-predicates": (
                "SELECT (?p AS ?predicate) (COUNT(?p) AS ?count) "
                "WHERE { ?s ?p ?o } "
                "GROUP BY ?p ORDER BY DESC(?count)"
            ),
            "all-graphs": (
                "SELECT ?g (COUNT(?g) AS ?count) "
                "WHERE { GRAPH ?g { ?s ?p ?o } } "
                "GROUP BY ?g ORDER BY DESC(?count)"
            ),
            "ten-random-triples": (
                "SELECT * WHERE { ?s ?p ?o } ORDER BY RAND() LIMIT 10"
            ),
        }

    def description(self) -> str:
        return "Send a query to a SPARQL endpoint"

    def should_have_qleverfile(self) -> bool:
        return False

    def relevant_qleverfile_arguments(self) -> dict[str, list[str]]:
        return {"server": ["host_name", "port", "access_token"]}

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "query",
            type=str,
            nargs="?",
            default="SELECT * WHERE { ?s ?p ?o } LIMIT 10",
            help="SPARQL query to send",
        )
        subparser.add_argument(
            "--predefined-query",
            type=str,
            choices=self.predefined_queries.keys(),
            help="Use a predefined query",
        )
        subparser.add_argument(
            "--pin-to-cache",
            action="store_true",
            default=False,
            help="Pin the query to the cache",
        )
        subparser.add_argument(
            "--sparql-endpoint", type=str, help="URL of the SPARQL endpoint"
        )
        subparser.add_argument(
            "--accept",
            type=str,
            choices=[
                "text/tab-separated-values",
                "text/csv",
                "application/sparql-results+json",
                "application/sparql-results+xml",
                "application/qlever-results+json",
                "application/octet-stream",
            ],
            default="text/tab-separated-values",
            help="Accept header for the SPARQL query",
        )
        subparser.add_argument(
            "--no-time",
            action="store_true",
            default=False,
            help="Do not print the (end-to-end) time taken",
        )

    def execute(self, args) -> bool:
        # Use a predefined query if requested.
        if args.predefined_query:
            args.query = self.predefined_queries[args.predefined_query]

        # When pinning to the cache, set `send=0` and request media type
        # `application/qlever-results+json` so that we get the result size.
        # Also, we need to provide the access token.
        if args.pin_to_cache:
            args.accept = "application/qlever-results+json"
            curl_cmd_additions = (
                f" --data pin-result=true --data send=0"
                f" --data access-token="
                f"{shlex.quote(args.access_token)}"
                f" | jq .resultsize | numfmt --grouping"
                f" | xargs -I {{}} printf"
                f' "Result pinned to cache,'
                f' number of rows: {{}}\\n"'
            )
        else:
            curl_cmd_additions = ""

        # Show what the command will do.
        sparql_endpoint = (
            args.sparql_endpoint
            if args.sparql_endpoint
            else f"{args.host_name}:{args.port}"
        )
        curl_cmd = (
            f"curl -s {sparql_endpoint}"
            f' -H "Accept: {args.accept}"'
            f" --data-urlencode query={shlex.quote(args.query)}"
            f"{curl_cmd_additions}"
        )
        self.show(curl_cmd, only_show=args.show)
        if args.show:
            return True

        # Launch query.
        try:
            start_time = time.time()
            run_command(curl_cmd, show_output=True)
            time_msecs = round(1000 * (time.time() - start_time))
            if not args.no_time and args.log_level != "NO_LOG":
                log.info("")
                log.info(
                    f"Query processing time (end-to-end):"
                    f" {time_msecs:,d} ms"
                )
        except Exception as e:
            if args.log_level == "DEBUG":
                traceback.print_exc()
            log.error(e)
            return False

        return True



from __future__ import annotations

import shlex
import time
import traceback

from qlever.command import QleverCommand
from qlever.log import log
from qlever.util import run_command


class UpdateCommand(QleverCommand):
    """
    Class for executing a SPARQL UPDATE against a SPARQL endpoint.

    The command accepts the update either directly on the command line or
    via a file path provided with --update-file.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return "Send an update to a SPARQL endpoint"

    def should_have_qleverfile(self) -> bool:
        return False

    def relevant_qleverfile_arguments(self) -> dict[str,list[str]]:
        return {"server": ["host_name", "port", "access_token"]}

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "update",
            type=str,
            nargs="?",
            default=None,
            help="SPARQL UPDATE to send (use --update-file to send from a file)",
        )
        subparser.add_argument(
            "--update-file",
            type=str,
            help="Path to a file containing the SPARQL UPDATE to send",
        )
        subparser.add_argument(
            "--sparql-endpoint", type=str, help="URL of the SPARQL endpoint"
        )

    def execute(self, args) -> bool:
        sparql_endpoint = (
            args.sparql_endpoint if args.sparql_endpoint else f"{args.host_name}:{args.port}"
        )

        curl_cmd = (
            f"curl -s {sparql_endpoint} -X POST "
            f"-H 'Authorization: Bearer {args.access_token}' "
            f"-H 'Content-Type: application/sparql-update' "
        )

        if args.update:
            curl_cmd += f"--data-binary {shlex.quote(args.update)}"
        elif args.update_file:
            curl_cmd += f"--data-binary @{shlex.quote(args.update_file)}"
        else:
            log.error("No SPARQL UPDATE provided. Pass it as an argument or via --update-file.")
            return False

        # Show and exit if requested
        self.show(curl_cmd, only_show=args.show)
        if args.show:
            return True

        # Execute update
        try:
            start_time = time.time()
            run_command(curl_cmd)
            time_msecs = round(1000 * (time.time() - start_time))
            if args.log_level != "NO_LOG":
                log.info("")
                log.info(
                    f"Update processing time (end-to-end): {time_msecs:,d} ms"
                )
        except Exception as e:
            if args.log_level == "DEBUG":
                traceback.print_exc()
            log.error(e)
            return False

        return True




from __future__ import annotations

import subprocess
import time

from qlever.command import QleverCommand
from qlever.commands.cache_stats import CacheStatsCommand
from qlever.commands.settings import SettingsCommand
from qlever.commands.status import StatusCommand
from qlever.commands.stop import StopCommand
from qlever.commands.warmup import WarmupCommand
from qlever.containerize import Containerize
from qlever.log import log
from qlever.qleverfile import Qleverfile
from qlever.util import binary_exists, is_qlever_server_alive, run_command


# Construct the command line based on the config file.
def construct_command(args) -> str:
    start_cmd = (
        f"{args.server_binary}"
        f" -i {args.name}"
        f" -j {args.num_threads}"
        f" -p {args.port}"
        f" -m {args.memory_for_queries}"
        f" -c {args.cache_max_size}"
        f" -e {args.cache_max_size_single_entry}"
        f" -k {args.cache_max_num_entries}"
    )

    if args.timeout:
        start_cmd += f" -s {args.timeout}"
    if args.access_token:
        start_cmd += f" -a {args.access_token}"
    if args.persist_updates:
        start_cmd += " --persist-updates"
    if args.only_pso_and_pos_permutations:
        start_cmd += " --only-pso-and-pos-permutations"
    if args.use_patterns == "no":
        start_cmd += " --no-patterns"
    if args.use_text_index == "yes":
        start_cmd += " -t"
    start_cmd += f" > {args.name}.server-log.txt 2>&1"
    return start_cmd


# Kill existing server on the same port. Trust that StopCommand() works?
# Maybe return StopCommand().execute(args) and handle it with a try except?
def kill_existing_server(args) -> bool:
    args.cmdline_regex = f"^ServerMain.* -p {args.port}"
    args.no_containers = True
    if not StopCommand().execute(args):
        log.error("Stopping the existing server failed")
        return False
    log.info("")
    return True


# Run the command in a container
def wrap_command_in_container(args, start_cmd) -> str:
    if not args.server_container:
        args.server_container = f"qlever.server.{args.name}"
    start_cmd = Containerize().containerize_command(
        start_cmd,
        args.system,
        "run -d --restart=unless-stopped",
        args.image,
        args.server_container,
        volumes=[("$(pwd)", "/index")],
        ports=[(args.port, args.port)],
        working_directory="/index",
    )
    return start_cmd


# Set the index description.
def set_index_description(access_arg, port, desc) -> bool:
    curl_cmd = (
        f"curl -Gs http://localhost:{port}/api"
        f' --data-urlencode "index-description={desc}"'
        f" {access_arg} > /dev/null"
    )
    log.debug(curl_cmd)
    try:
        run_command(curl_cmd)
    except Exception as e:
        log.error(f"Setting the index description failed ({e})")
        return False
    return True


# Set the text description.
def set_text_description(access_arg, port, text_desc) -> bool:
    curl_cmd = (
        f"curl -Gs http://localhost:{port}/api"
        f' --data-urlencode "text-description={text_desc}"'
        f" {access_arg} > /dev/null"
    )
    log.debug(curl_cmd)
    try:
        run_command(curl_cmd)
    except Exception as e:
        log.error(f"Setting the text description failed ({e})")
        return False
    return True


class StartCommand(QleverCommand):
    """
    Class for executing the `start` command.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return (
            "Start the QLever server (requires that you have built "
            "an index with `qlever index` before)"
        )

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str, list[str]]:
        return {
            "data": ["name", "description", "text_description"],
            "server": [
                "server_binary",
                "host_name",
                "port",
                "access_token",
                "memory_for_queries",
                "cache_max_size",
                "cache_max_size_single_entry",
                "cache_max_num_entries",
                "num_threads",
                "timeout",
                "persist_updates",
                "only_pso_and_pos_permutations",
                "use_patterns",
                "use_text_index",
                "warmup_cmd",
            ],
            "runtime": ["system", "image", "server_container"],
        }

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "--kill-existing-with-same-port",
            action="store_true",
            default=False,
            help="If a QLever server is already running "
            "on the same port, kill it before "
            "starting a new server",
        )
        subparser.add_argument(
            "--no-warmup",
            action="store_true",
            default=False,
            help="Do not execute the warmup command",
        )
        subparser.add_argument(
            "--run-in-foreground",
            action="store_true",
            default=False,
            help="Run the server in the foreground "
            "(default: run in the background with `nohup`)",
        )
        subparser.add_argument(
            "runtime_parameters",
            nargs="*",
            help="Space-separated list of runtime parameters to set "
            "(in the form `key=value`) once the server is running",
        ).completer = lambda **kwargs: [
            f"{key}=" for key in Qleverfile.SERVER_RUNTIME_PARAMETERS
        ]

    def execute(self, args) -> bool:
        # Set the endpoint URL.
        args.endpoint_url = f"http://{args.host_name}:{args.port}"

        # Kill existing server with the same name if so desired.
        #
        # TODO: This is currently disabled because I never used it once over
        # the past weeks and it is not clear to me what the use case is.
        if False:  # or args.kill_existing_with_same_name:
            args.cmdline_regex = f"^ServerMain.* -i {args.name}"
            args.no_containers = True
            StopCommand().execute(args)
            log.info("")

        # Kill existing server on the same port if so desired.
        if args.kill_existing_with_same_port:
            if args.kill_existing_with_same_port and not kill_existing_server(
                args
            ):
                return False

        # Construct the command line based on the config file.
        start_cmd = construct_command(args)

        # Run the command in a container (if so desired). Otherwise run with
        # `nohup` so that it keeps running after the shell is closed. With
        # `--run-in-foreground`, run the server in the foreground.
        if args.system in Containerize.supported_systems():
            start_cmd = wrap_command_in_container(args, start_cmd)
        elif args.run_in_foreground:
            start_cmd = f"{start_cmd}"
        else:
            start_cmd = f"nohup {start_cmd} &"

        # Show the command line.
        self.show(start_cmd, only_show=args.show)
        if args.show:
            if args.runtime_parameters:
                log.info("")
                SettingsCommand().execute(args)
            return True

        # When running natively, check if the binary exists and works.
        if args.system == "native":
            if not binary_exists(args.server_binary, "server-binary"):
                return False

        # Check if a QLever server is already running on this port.
        if is_qlever_server_alive(args.endpoint_url):
            log.error(f"QLever server already running on {args.endpoint_url}")
            log.info("")
            log.info(
                "To kill the existing server, use `qlever stop` "
                "or `qlever start` with option "
                "--kill-existing-with-same-port`"
            )

            # Show output of status command.
            args.cmdline_regex = f"^ServerMain.* -p *{args.port}"
            log.info("")
            StatusCommand().execute(args)
            return False

        # Remove already existing container.
        if (
            args.system in Containerize.supported_systems()
            and args.kill_existing_with_same_port
        ):
            try:
                run_command(f"{args.system} rm -f {args.server_container}")
            except Exception as e:
                log.error(f"Removing existing container failed: {e}")
                return False

        # Check if another process is already listening.
        # if self.net_connections_enabled:
        #     if port in [conn.laddr.port for conn
        #                 in psutil.net_connections()]:
        #         log.error(f"Port {port} is already in use by another process"
        #                   f" (use `lsof -i :{port}` to find out which one)")
        #         return False

        # Execute the command line.
        try:
            process = run_command(
                start_cmd,
                use_popen=args.run_in_foreground,
            )
        except Exception as e:
            log.error(f"Starting the QLever server failed ({e})")
            return False

        # Tail the server log until the server is ready (note that the `exec`
        # is important to make sure that the tail process is killed and not
        # just the bash process).
        if args.run_in_foreground:
            log.info(
                f"Follow {args.name}.server-log.txt as long as the server is"
                f" running (Ctrl-C stops the server)"
            )
        else:
            log.info(
                f"Follow {args.name}.server-log.txt until the server is ready"
                f" (Ctrl-C stops following the log, but NOT the server)"
            )
        log.info("")
        tail_cmd = f"exec tail -f {args.name}.server-log.txt"
        tail_proc = subprocess.Popen(tail_cmd, shell=True)
        while not is_qlever_server_alive(args.endpoint_url):
            time.sleep(1)

        # Set the description for the index and text.
        access_arg = f'--data-urlencode "access-token={args.access_token}"'
        if args.description:
            ret = set_index_description(
                access_arg, args.port, args.description
            )
            if not ret:
                return False
        if args.text_description:
            ret = set_text_description(
                access_arg, args.port, args.text_description
            )
            if not ret:
                return False

        # Kill the tail process. NOTE: `tail_proc.kill()` does not work.
        if not args.run_in_foreground:
            tail_proc.terminate()

        # Execute the warmup command.
        if args.warmup_cmd and not args.no_warmup:
            log.info("")
            if not WarmupCommand().execute(args):
                log.error("Warmup failed")
                return False

        # Show cache stats.
        if not args.run_in_foreground:
            log.info("")
            args.detailed = False
            args.sparql_endpoint = None
            CacheStatsCommand().execute(args)

        # Apply settings if any.
        if args.runtime_parameters:
            log.info("")
            SettingsCommand().execute(args)

        # With `--run-in-foreground`, wait until the server is stopped.
        if args.run_in_foreground:
            try:
                process.wait()
            except KeyboardInterrupt:
                process.terminate()
            tail_proc.terminate()

        return True



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK

# Copyright 2024, University of Freiburg,
# Chair of Algorithms and Data Structures
# Author: Hannah Bast <bast@cs.uni-freiburg.de>

from __future__ import annotations

import re
import traceback

from termcolor import colored

from qlever import command_objects, script_name
from qlever.config import ConfigException, QleverConfig
from qlever.log import log, log_levels


def main():
    # Parse the command line arguments and read the Qleverfile.
    try:
        qlever_config = QleverConfig()
        args = qlever_config.parse_args()
    except ConfigException as e:
        log.error(e)
        log.info("")
        log.info(traceback.format_exc())
        exit(1)

    # Execute the command.
    command_object = command_objects[args.command]
    log.setLevel(log_levels[args.log_level])
    try:
        log.info("")
        log.info(colored(f"Command: {args.command}", attrs=["bold"]))
        log.info("")
        command_successful = command_object.execute(args)
        log.info("")
        if not command_successful:
            exit(1)
    except KeyboardInterrupt:
        log.warn("\rCtrl-C pressed, exiting ...")
        log.info("")
        exit(1)
    except Exception as e:
        # Check if it's a certain kind of `AttributeError` and give a hint in
        # that case.
        log.debug(
            "Command failed with exception, full traceback: "
            f"{traceback.format_exc()}"
        )
        match_error = re.search(r"object has no attribute '(.+)'", str(e))
        match_trace = re.search(
            rf"({script_name}/commands/.+\.py)\", line (\d+)",
            traceback.format_exc(),
        )
        if isinstance(e, AttributeError) and match_error and match_trace:
            attribute = match_error.group(1)
            trace_command = match_trace.group(1)
            trace_line = match_trace.group(2)
            log.error(f"{e} in `{trace_command}` at line {trace_line}")
            log.info("")
            log.info(
                f"Likely cause: you used `args.{attribute}`, but it was "
                f"neither defined in `relevant_qleverfile_arguments` "
                f"nor in `additional_arguments`"
            )
            log.info("")
            log.info(
                f"If you did not implement `{trace_command}` yourself, "
                f"please report this issue"
            )
            log.info("")
        else:
            log.error(f"An unexpected error occurred: {e}")
            log.info("")
            log.info(traceback.format_exc())
        exit(1)



from __future__ import annotations

import json
import logging
import os
import re
import signal
import time
from datetime import datetime, timezone
from enum import Enum, auto

import rdflib.term
import requests_sse
from rdflib import Graph
from termcolor import colored
from tqdm.contrib.logging import tqdm_logging_redirect

from qlever.command import QleverCommand
from qlever.log import log
from qlever.util import run_command


# Monkey patch `rdflib.term._castLexicalToPython` to avoid casting of literals
# to Python types. We do not need it (all we want it convert Turtle to N-Triples),
# and we can speed up parsing by a factor of about 2.
def custom_cast_lexical_to_python(lexical, datatype):
    return None  # Your desired behavior


rdflib.term._castLexicalToPython = custom_cast_lexical_to_python


class UpdateWikidataCommand(QleverCommand):
    """
    Class for executing the `update` command.
    """

    def __init__(self):
        # SPARQL query to get the date until which the updates of the
        # SPARQL endpoint are complete.
        self.sparql_updates_complete_until_query = (
            "PREFIX wikibase: <http://wikiba.se/ontology#> "
            "PREFIX schema: <http://schema.org/> "
            "SELECT * WHERE { "
            "{ SELECT (MIN(?date_modified) AS ?updates_complete_until) { "
            "wikibase:Dump schema:dateModified ?date_modified } } "
            "UNION { wikibase:Dump wikibase:updatesCompleteUntil ?updates_complete_until } "
            "} ORDER BY DESC(?updates_complete_until) LIMIT 1"
        )
        # URL of the Wikidata SSE stream.
        self.wikidata_update_stream_url = (
            "https://stream.wikimedia.org/v2/"
            "stream/rdf-streaming-updater.mutation.v2"
        )
        # Remember if Ctrl+C was pressed, so we can handle it gracefully.
        self.ctrl_c_pressed = False
        # Set to `True` when finished.
        self.finished = False

    def description(self) -> str:
        return "Update from given SSE stream"

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str, list[str]]:
        return {"server": ["host_name", "port", "access_token"]}

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "sse_stream_url",
            nargs="?",
            type=str,
            default=self.wikidata_update_stream_url,
            help="URL of the SSE stream to update from",
        )
        subparser.add_argument(
            "--batch-size",
            type=int,
            default=100000,
            help="Group this many messages together into one update "
            "(default: one update for each message); NOTE: this simply "
            "concatenates the `rdf_added_data` and `rdf_deleted_data` fields, "
            "which is not 100%% correct; as soon as chaining is supported, "
            "this will be fixed",
        )
        subparser.add_argument(
            "--lag-seconds",
            type=int,
            default=1,
            help="When a message is encountered that is within this many "
            "seconds of the current time, finish the current batch "
            "(and show a warning that this happened)",
        )
        subparser.add_argument(
            "--since",
            type=str,
            help="Consume stream messages since this date "
            "(default: determine automatically from the SPARQL endpoint)",
        )
        subparser.add_argument(
            "--until",
            type=str,
            help="Stop consuming stream messages when reaching this date "
            "(default: continue indefinitely)",
        )
        subparser.add_argument(
            "--offset",
            type=int,
            help="Consume stream messages starting from this offset "
            "(default: not set)",
        )
        subparser.add_argument(
            "--topic",
            type=str,
            choices=[
                "eqiad.rdf-streaming-updater.mutation",
                "codfw.rdf-streaming-updater.mutation",
            ],
            default="eqiad.rdf-streaming-updater.mutation",
            help="The topic to consume from the SSE stream (default: "
            "eqiad.rdf-streaming-updater.mutation)",
        )
        subparser.add_argument(
            "--partition",
            type=int,
            default=0,
            help="The partition to consume from the SSE stream (default: 0)",
        )
        subparser.add_argument(
            "--min-or-max-date",
            choices=["min", "max"],
            default="max",
            help="Use the minimum or maximum date of the batch for the "
            "`updatesCompleteUntil` property (default: maximum)",
        )
        subparser.add_argument(
            "--wait-between-batches",
            type=int,
            default=300,
            help="Wait this many seconds between batches that were "
            "finished due to a message that is within `lag_seconds` of "
            "the current time (default: 300s)",
        )
        subparser.add_argument(
            "--num-messages",
            type=int,
            help="Process exactly this many messages and then exit "
            "(default: no bound on the number of messages)",
        )
        subparser.add_argument(
            "--verbose",
            choices=["no", "yes"],
            default="yes",
            help='Verbose logging, "yes" or "no" (default: "yes")',
        )
        subparser.add_argument(
            "--use-cached-sparql-queries",
            action="store_true",
            help="Use cached SPARQL query files if they exist with matching "
            "offset and target batch size (default: off)",
        )
        subparser.add_argument(
            "--check-offset-before-each-batch",
            choices=["yes", "no"],
            default="yes",
            help="Before each batch, verify that the stream offset matches the "
            "stored offset in the knowledge base (default: yes)",
        )

    # Handle Ctrl+C gracefully by finishing the current batch and then exiting.
    def handle_ctrl_c(self, signal_received, frame):
        if self.ctrl_c_pressed:
            log.warn("\rCtrl+C pressed again, watch your blood pressure")
        else:
            self.ctrl_c_pressed = True

    def execute(self, args) -> bool:
        # cURL command to get the date until which the updates of the
        # SPARQL endpoint are complete.
        sparql_endpoint = f"http://{args.host_name}:{args.port}"
        curl_cmd_updates_complete_until = (
            f"curl -s {sparql_endpoint}"
            f' -H "Accept: text/csv"'
            f' -H "Content-type: application/sparql-query"'
            f' --data "{self.sparql_updates_complete_until_query}"'
        )

        # Construct the command and show it.
        cmd_description = []
        if args.since:
            cmd_description.append(f"SINCE={args.since}")
        else:
            cmd_description.append(
                f"SINCE=$({curl_cmd_updates_complete_until} | sed 1d)"
            )
        if args.until:
            cmd_description.append(f"UNTIL={args.until}")
        cmd_description.append(
            f"Process SSE stream from {args.sse_stream_url} "
            f"in batches of up to {args.batch_size:,} messages "
        )
        self.show("\n".join(cmd_description), only_show=args.show)
        if args.show:
            return True

        # Compute the `since` date if not given.
        if args.since:
            since = args.since
        else:
            try:
                since = run_command(
                    f"{curl_cmd_updates_complete_until} | sed 1d",
                    return_output=True,
                ).strip()
            except Exception as e:
                log.error(
                    f"Error running `{curl_cmd_updates_complete_until}`: {e}"
                )
                return False

        # Special handling of Ctrl+C, see `handle_ctrl_c` above.
        signal.signal(signal.SIGINT, self.handle_ctrl_c)
        log.warn("Press Ctrl+C to finish and exit gracefully")
        log.info("")

        # If --offset is not provided, determine the offset by reading a single
        # message from the SSE stream using the `since` date.
        if not args.offset:
            try:
                source = requests_sse.EventSource(
                    args.sse_stream_url,
                    params={"since": since},
                    headers={
                        "Accept": "text/event-stream",
                        "User-Agent": "qlever update-wikidata",
                    },
                )
                source.connect()
                for event in source:
                    if event.type == "message" and event.data:
                        event_data = json.loads(event.data)
                        topic = event_data.get("meta").get("topic")
                        if topic == args.topic:
                            args.offset = event_data.get("meta").get("offset")
                            log.debug(
                                f"Determined offset from date: "
                                f"{since} -> {args.offset}"
                            )
                            break
                source.close()
            except Exception as e:
                log.error(f"Error determining offset from stream: {e}")
                return False

        # Initialize all the statistics variables.
        batch_count = 0
        total_num_messages = 0
        total_num_ops = 0
        total_time_s = 0
        start_time = time.perf_counter()
        wait_before_next_batch = False
        event_id_for_next_batch = (
            [
                {
                    "topic": args.topic,
                    "partition": args.partition,
                    "offset": args.offset,
                }
            ]
            if args.offset
            else None
        )

        # Track whether this is the first batch (to skip offset check)
        first_batch = True

        # Main event loop: Either resume from `event_id_for_next_batch` (if set),
        # or start a new connection to `args.sse_stream_url` (with URL
        # parameter `?since=`).
        while True:
            # Optionally wait before processing the next batch (make sure that
            # the wait is interruptible by Ctrl+C).
            if wait_before_next_batch:
                log.info(
                    f"Waiting {args.wait_between_batches} "
                    f"second{'s' if args.wait_between_batches > 1 else ''} "
                    f"before processing the next batch"
                )
                log.info("")
                wait_before_next_batch = False
                for _ in range(args.wait_between_batches):
                    if self.ctrl_c_pressed:
                        break
                    time.sleep(1)
            if self.ctrl_c_pressed:
                log.warn(
                    "\rCtrl+C pressed while waiting in between batches, "
                    "exiting"
                )
                break

            # Start stream from either `event_id_for_next_batch` or `since`.
            # We'll extract the offset for first_offset_in_batch later.
            if event_id_for_next_batch:
                event_id_json = json.dumps(event_id_for_next_batch)
                if args.verbose == "yes":
                    log.info(
                        colored(
                            f"Consuming stream from event ID: {event_id_json}",
                            attrs=["dark"],
                        )
                    )
                source = requests_sse.EventSource(
                    args.sse_stream_url,
                    headers={
                        "Accept": "text/event-stream",
                        "User-Agent": "qlever update-wikidata",
                        "Last-Event-ID": event_id_json,
                    },
                )
            else:
                if args.verbose == "yes":
                    log.info(
                        colored(
                            f"Consuming stream from date: {since}",
                            attrs=["dark"],
                        )
                    )
                source = requests_sse.EventSource(
                    args.sse_stream_url,
                    params={"since": since},
                    headers={
                        "Accept": "text/event-stream",
                        "User-Agent": "qlever update-wikidata",
                    },
                )
            # Try 10 times to connect to the source, with exponential backoff.
            # If we cannot connect after that, give up.
            wait_between_connect_attempts_s = 1
            source_connected = False
            while True:
                try:
                    source.connect()
                    source_connected = True
                    break
                except Exception:
                    log.error(
                        f"Error connecting to stream, waiting "
                        f"{wait_between_connect_attempts_s}s before retrying"
                    )
                    time.sleep(wait_between_connect_attempts_s)
                    wait_between_connect_attempts_s *= 2
                    if wait_between_connect_attempts_s > 512:
                        log.error("Giving up connecting to SSE source, exit")
                        break
            if not source_connected:
                break

            # Next comes the inner loop, which processes exactly one "batch" of
            # messages. The batch is completed (simply using `break`) when either
            # `args.batch_size` messages have been processed, or when one of a
            # variety of conditions occur (Ctrl+C pressed, message within
            # `args.lag_seconds` of current time, delete operation followed by
            # insert of triple with that entity as subject).

            # Initialize all the batch variables.
            current_batch_size = 0
            # Extract the offset from the event ID to use as the starting offset
            # for this batch. This is set before processing any messages.
            if event_id_for_next_batch:
                first_offset_in_batch = event_id_for_next_batch[0]["offset"]
                event_id_for_next_batch = None
            else:
                # This should not happen since we now always determine the offset
                # before starting, but keep as fallback
                first_offset_in_batch = None

            # Check that the stream offset matches the stored offset in the KB
            # Skip this check on the first batch (when using --offset to resume)
            if (
                args.check_offset_before_each_batch == "yes"
                and not first_batch
                and first_offset_in_batch is not None
            ):
                sparql_query_offset = (
                    "PREFIX wikibase: <http://wikiba.se/ontology#> "
                    "SELECT (MAX(?offset) AS ?maxOffset) WHERE { "
                    "<http://wikiba.se/ontology#Dump> "
                    "wikibase:updateStreamNextOffset ?offset "
                    "}"
                )
                curl_cmd_check_offset = (
                    f"curl -s {sparql_endpoint}"
                    f' -H "Accept: text/csv"'
                    f' -H "Content-type: application/sparql-query"'
                    f' --data "{sparql_query_offset}"'
                )
                try:
                    result = run_command(
                        f"{curl_cmd_check_offset} | sed 1d",
                        return_output=True,
                    ).strip()
                    if not result:
                        log.error(
                            "Failed to retrieve stored offset from knowledge base: "
                            "query returned no results. This might be the first update, "
                            "or the offset triple is missing."
                        )
                        return False
                    stored_offset = int(result.strip('"'))
                    if stored_offset != first_offset_in_batch:
                        log.error(
                            f"Offset mismatch: stream offset is {first_offset_in_batch}, "
                            f"but stored offset in knowledge base is {stored_offset}. "
                            f"This indicates that updates may have been applied "
                            f"out of order or some updates are missing."
                        )
                        return False
                except Exception as e:
                    log.error(
                        f"Failed to retrieve or verify stored offset from knowledge base: {e}. "
                        f"Cannot safely proceed with updates."
                    )
                    return False

            date_list = []
            delete_entity_ids = set()
            delta_to_now_list = []
            batch_assembly_start_time = time.perf_counter()
            insert_triples = set()
            delete_triples = set()

            # Check if we can use a cached SPARQL query file
            use_cached_file = False
            cached_file_name = None
            cached_meta_file_name = None
            cached_date_range = None
            if (
                args.use_cached_sparql_queries
                and first_offset_in_batch is not None
            ):
                cached_file_name = (
                    f"update.{first_offset_in_batch}.{args.batch_size}.sparql"
                )
                cached_meta_file_name = (
                    f"update.{first_offset_in_batch}.{args.batch_size}.meta"
                )
                if os.path.exists(cached_file_name):
                    use_cached_file = True
                    # Try to read metadata file for date range
                    if os.path.exists(cached_meta_file_name):
                        try:
                            with open(cached_meta_file_name, "r") as f:
                                cached_date_range = f.read().strip()
                        except Exception:
                            cached_date_range = None

                    if args.verbose == "yes":
                        log_msg = f"Using cached SPARQL query file: {cached_file_name}"
                        if cached_date_range:
                            log_msg += f" [date range: {cached_date_range}]"
                        log.info(colored(log_msg, "cyan"))

            # Process one event at a time (unless using cached file).
            if not use_cached_file:
                with tqdm_logging_redirect(
                    loggers=[logging.getLogger("qlever")],
                    desc="Batch",
                    total=args.batch_size,
                    leave=False,
                    bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt}{postfix}",
                ) as pbar:
                    for event in source:
                        # Skip events that are not of type `message` (should not
                        # happen), have no field `data` (should not happen either), or
                        # where the topic is not in `args.topics` (one topic by itself
                        # should provide all relevant updates).
                        if event.type != "message" or not event.data:
                            continue
                        event_data = json.loads(event.data)
                        topic = event_data.get("meta").get("topic")
                        if topic != args.topic:
                            continue

                        try:
                            # Extract offset, topic, and partition from the message metadata
                            # to construct a precise event ID for resuming.
                            meta = event_data.get("meta")
                            offset = meta.get("offset")
                            topic = meta.get("topic")
                            partition = meta.get("partition")

                            # Get the date (rounded *down* to seconds).
                            date = meta.get("dt")
                            date = re.sub(r"\.\d*Z$", "Z", date)

                            # Get the other relevant fields from the message.
                            entity_id = event_data.get("entity_id")
                            operation = event_data.get("operation")
                            rdf_added_data = event_data.get("rdf_added_data")
                            rdf_deleted_data = event_data.get(
                                "rdf_deleted_data"
                            )
                            rdf_linked_shared_data = event_data.get(
                                "rdf_linked_shared_data"
                            )
                            rdf_unlinked_shared_data = event_data.get(
                                "rdf_unlinked_shared_data"
                            )

                            # Check batch completion conditions BEFORE processing the
                            # data of this message. If any of the conditions is met,
                            # we finish the batch and resume from the LAST PROCESSED
                            # message (not the current one that triggered the break).
                            #
                            # NOTE: We will update event_id_for_next_batch AFTER
                            # successfully processing each message (see below), so that
                            # when we break, it contains the last processed event ID.
                            since = None

                            # Condition 1: Delete followed by insert for same entity.
                            operation_adds_data = (
                                rdf_added_data is not None
                                or rdf_linked_shared_data is not None
                            )
                            if (
                                operation_adds_data
                                and entity_id in delete_entity_ids
                            ):
                                if args.verbose == "yes":
                                    log.warn(
                                        f"Encountered operation that adds data for "
                                        f"an entity ID ({entity_id}) that was deleted "
                                        f"earlier in this batch; finishing batch and "
                                        f"resuming from this message in the next batch"
                                    )
                                break

                            # Condition 2: Batch size or limit on number of
                            # messages reached.
                            if current_batch_size >= args.batch_size or (
                                args.num_messages is not None
                                and total_num_messages >= args.num_messages
                            ):
                                break

                            # Condition 3: Message close to current time.
                            date_obj = datetime.strptime(
                                date, "%Y-%m-%dT%H:%M:%SZ"
                            ).replace(tzinfo=timezone.utc)
                            date_as_epoch_s = date_obj.timestamp()

                            now_as_epoch_s = time.time()
                            delta_to_now_s = now_as_epoch_s - date_as_epoch_s
                            if (
                                delta_to_now_s < args.lag_seconds
                                and current_batch_size > 0
                            ):
                                if args.verbose == "yes":
                                    log.warn(
                                        f"Encountered message with date {date}, which is within "
                                        f"{args.lag_seconds} "
                                        f"second{'s' if args.lag_seconds > 1 else ''} "
                                        f"of the current time, finishing the current batch"
                                    )
                                wait_before_next_batch = (
                                    args.wait_between_batches is not None
                                    and args.wait_between_batches > 0
                                )
                                break

                            # Condition 4: Reached `--until` date and at least one
                            # message was processed.
                            if (
                                args.until
                                and date >= args.until
                                and current_batch_size > 0
                            ):
                                log.warn(
                                    f"Reached --until date {args.until} "
                                    f"(message date: {date}), that's it folks"
                                )
                                self.finished = True
                                break

                            # Delete operations are postponed until the end of the
                            # batch, so remember the entity ID here.
                            if operation == "delete":
                                delete_entity_ids.add(entity_id)

                            # Replace each occurrence of `\\` by `\u005C\u005C`
                            # (which is twice the Unicode for backslash).
                            #
                            # NOTE: Strictly speaking, it would be enough to do
                            # this for two backslashes followed by a `u`, but
                            # doing it for all double backslashes does not
                            # harm. When parsing a SPARQL query, then according
                            # to the standar, first all occurrences of `\uxxxx`
                            # (where `xxxx` are four hex digits) are replaced
                            # by the corresponding Unicode character. That is a
                            # problem when `\\uxxxx` occurs in a literal,
                            # because then it would be replaced by `\` followed
                            # by the Unicode character, which is invalied
                            # SPARQL. The subsitution avoids that problem.
                            def node_to_sparql(node: rdflib.term.Node) -> str:
                                return node.n3().replace(
                                    "\\\\", "\\u005C\\u005C"
                                )

                            # Process the to-be-deleted triples.
                            for rdf_to_be_deleted in (
                                rdf_deleted_data,
                                rdf_unlinked_shared_data,
                            ):
                                if rdf_to_be_deleted is not None:
                                    try:
                                        rdf_to_be_deleted_data = (
                                            rdf_to_be_deleted.get("data")
                                        )
                                        graph = Graph()
                                        log.debug(
                                            f"RDF to_be_deleted data: {rdf_to_be_deleted_data}"
                                        )
                                        graph.parse(
                                            data=rdf_to_be_deleted_data,
                                            format="turtle",
                                        )
                                        for s, p, o in graph:
                                            triple = f"{s.n3()} {p.n3()} {node_to_sparql(o)}"
                                            # NOTE: In case there was a previous `insert` of that
                                            # triple, it is safe to remove that `insert`, but not
                                            # the `delete` (in case the triple is contained in the
                                            # original data).
                                            if triple in insert_triples:
                                                insert_triples.remove(triple)
                                            delete_triples.add(triple)
                                    except Exception as e:
                                        log.error(
                                            f"Error reading `rdf_to_be_deleted_data`: {e}"
                                        )
                                        return False

                            # Process the to-be-added triples.
                            for rdf_to_be_added in (
                                rdf_added_data,
                                rdf_linked_shared_data,
                            ):
                                if rdf_to_be_added is not None:
                                    try:
                                        rdf_to_be_added_data = (
                                            rdf_to_be_added.get("data")
                                        )
                                        graph = Graph()
                                        log.debug(
                                            "RDF to be added data: {rdf_to_be_added_data}"
                                        )
                                        graph.parse(
                                            data=rdf_to_be_added_data,
                                            format="turtle",
                                        )
                                        for s, p, o in graph:
                                            triple = f"{s.n3()} {p.n3()} {node_to_sparql(o)}"
                                            # NOTE: In case there was a previous `delete` of that
                                            # triple, it is safe to remove that `delete`, but not
                                            # the `insert` (in case the triple is not contained in
                                            # the original data).
                                            if triple in delete_triples:
                                                delete_triples.remove(triple)
                                            insert_triples.add(triple)
                                    except Exception as e:
                                        log.error(
                                            f"Error reading `rdf_to_be_added_data`: {e}"
                                        )
                                        return False

                        except Exception as e:
                            log.error(f"Error reading data from message: {e}")
                            log.info(event)
                            continue

                        # Message was successfully processed, update batch tracking
                        current_batch_size += 1
                        total_num_messages += 1
                        pbar_update_frequency = 100
                        if (current_batch_size % pbar_update_frequency) == 0:
                            pbar.set_postfix(
                                {
                                    "Time": date_obj.strftime(
                                        "%Y-%m-%d %H:%M:%S"
                                    )
                                }
                            )
                            pbar.update(pbar_update_frequency)
                        log.debug(
                            f"DATE: {date_as_epoch_s:.0f} [{date}], "
                            f"NOW: {now_as_epoch_s:.0f}, "
                            f"DELTA: {now_as_epoch_s - date_as_epoch_s:.0f}"
                        )
                        date_list.append(date)
                        delta_to_now_list.append(delta_to_now_s)

                        # Update the event ID for the next batch. We increment the
                        # offset by 1 so that the next batch starts with the next
                        # message (not re-processing the current one).
                        event_id_for_next_batch = [
                            {
                                "topic": topic,
                                "partition": partition,
                                "offset": offset + 1,
                            }
                        ]

                        # Ctrl+C finishes the current batch (this should come at the
                        # end of the inner event loop so that always at least one
                        # message is processed).
                        if self.ctrl_c_pressed:
                            log.warn(
                                "\rCtrl+C pressed while processing a batch, "
                                "finishing it and exiting"
                            )
                            break
            else:
                # Using cached file - set batch size and calculate next offset
                current_batch_size = args.batch_size
                total_num_messages += current_batch_size
                event_id_for_next_batch = [
                    {
                        "topic": args.topic,
                        "partition": args.partition,
                        "offset": first_offset_in_batch + current_batch_size,
                    }
                ]

            # Process the current batch of messages (or skip if using cached).
            batch_count += 1
            if not use_cached_file:
                batch_assembly_end_time = time.perf_counter()
                batch_assembly_time_ms = int(
                    1000
                    * (batch_assembly_end_time - batch_assembly_start_time)
                )
                date_list.sort()
                delta_to_now_list.sort()
                min_delta_to_now_s = delta_to_now_list[0]
                if min_delta_to_now_s < 10:
                    min_delta_to_now_s = f"{min_delta_to_now_s:.1f}"
                else:
                    min_delta_to_now_s = f"{int(min_delta_to_now_s):,}"
                log.info(
                    f"Assembled batch #{batch_count}, "
                    f"#messages: {current_batch_size:2,}, "
                    f"date range: {date_list[0]} - {date_list[-1]}  "
                    f"[assembly time: {batch_assembly_time_ms:3,}ms, "
                    f"min delta to NOW: {min_delta_to_now_s}s]"
                )

                # Add the min and max date of the batch to `insert_triples`.
                #
                # NOTE: The min date means that we have *all* updates until that
                # date. The max date is the date of the latest update we have seen.
                # However, there may still be earlier updates that we have not seen
                # yet. Wikidata uses `schema:dateModified` for the latter semantics,
                # so we use it here as well. For the other semantics, we invent
                # a new property `wikibase:updatesCompleteUntil`.
                insert_triples.add(
                    f"<http://wikiba.se/ontology#Dump> "
                    f"<http://schema.org/dateModified> "
                    f'"{date_list[-1]}"^^<http://www.w3.org/2001/XMLSchema#dateTime>'
                )
                updates_complete_until = (
                    date_list[-1]
                    if args.min_or_max_date == "max"
                    else date_list[0]
                )
                insert_triples.add(
                    f"<http://wikiba.se/ontology#Dump> "
                    f"<http://wikiba.se/ontology#updatesCompleteUntil> "
                    f'"{updates_complete_until}"'
                    f"^^<http://www.w3.org/2001/XMLSchema#dateTime>"
                )
                insert_triples.add(
                    "<http://wikiba.se/ontology#Dump> "
                    "<http://wikiba.se/ontology#updateStreamNextOffset> "
                    f'"{event_id_for_next_batch[0]["offset"]}"'
                )

                # Construct UPDATE operation.
                delete_block = " . \n  ".join(delete_triples)
                insert_block = " . \n  ".join(insert_triples)
                delete_insert_operation = (
                    f"DELETE {{\n  {delete_block} \n}} "
                    f"INSERT {{\n  {insert_block} \n}} "
                    f"WHERE {{ }}\n"
                )

                # If `delete_entity_ids` is non-empty, add a `DELETE WHERE`
                # operation that deletes all triples that are associated with only
                # those entities.
                delete_entity_ids_as_values = " ".join(
                    [f"wd:{qid}" for qid in delete_entity_ids]
                )
                if len(delete_entity_ids) > 0:
                    delete_where_operation = (
                        f"PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>\n"
                        f"PREFIX wikibase: <http://wikiba.se/ontology#>\n"
                        f"PREFIX wd: <http://www.wikidata.org/entity/>\n"
                        f"DELETE {{\n"
                        f"  ?s ?p ?o .\n"
                        f"}} WHERE {{\n"
                        f"  {{\n"
                        f"    VALUES ?s {{ {delete_entity_ids_as_values} }}\n"
                        f"    ?s ?p ?o .\n"
                        f"  }} UNION {{\n"
                        f"    VALUES ?_1 {{ {delete_entity_ids_as_values} }}\n"
                        f"    ?_1 ?_2 ?s .\n"
                        f"    ?s ?p ?o .\n"
                        f"    ?s rdf:type wikibase:Statement .\n"
                        f"  }}\n"
                        f"}}\n"
                    )
                    delete_insert_operation += ";\n" + delete_where_operation

            # Construct curl command. For batch size 1, send the operation via
            # `--data-urlencode`, otherwise write to file and send via `--data-binary`.
            curl_cmd = (
                f"curl -s -X POST"
                f' "{sparql_endpoint}?access-token={args.access_token}"'
                f" -H 'Content-Type: application/sparql-update'"
            )
            if use_cached_file:
                # Use the cached file instead of writing a new one
                update_arg_file_name = cached_file_name
            else:
                # Write the constructed SPARQL update to a file
                update_arg_file_name = f"update.{first_offset_in_batch}.{current_batch_size}.sparql"
                with open(update_arg_file_name, "w") as f:
                    f.write(delete_insert_operation)
                # Write metadata file with date range
                meta_file_name = (
                    f"update.{first_offset_in_batch}.{current_batch_size}.meta"
                )
                with open(meta_file_name, "w") as f:
                    f.write(f"{date_list[0]} - {date_list[-1]}")
            curl_cmd += f" --data-binary @{update_arg_file_name}"
            if args.verbose == "yes":
                log.info(colored(curl_cmd, "blue"))

            # Run it (using `curl` for batch size up to 1000, otherwise
            # `requests`).
            try:
                result = run_command(curl_cmd, return_output=True)
                result_file_name = f"update.{first_offset_in_batch}.{current_batch_size}.result"
                with open(result_file_name, "w") as f:
                    f.write(result)
            except Exception as e:
                log.warn(f"Error running `curl` command: {e}")
                log.info("")
                continue

            # Results should be a JSON, parse it.
            try:
                result = json.loads(result)
            except Exception as e:
                log.error(
                    f"Error parsing JSON result: {e}"
                    f", the first 1000 characters are:"
                )
                log.info(result[:1000])
                log.info("")
                continue

            # Check if the result contains a QLever exception.
            if "exception" in result:
                error_msg = result["exception"]
                log.error(f"QLever exception: {error_msg}")
                log.info("")
                continue

            # Helper function for getting the value of `stats["time"][...]`
            # without the "ms" suffix. If the extraction fails, return 0

            # (and optionally log the failure).
            class FailureMode(Enum):
                LOG_ERROR = auto()
                SILENTLY_RETURN_ZERO = auto()
                THROW_EXCEPTION = auto()

            def get_time_ms(
                stats, *keys: str, failure_mode=FailureMode.LOG_ERROR
            ) -> int:
                try:
                    value = stats["time"]
                    for key in keys:
                        value = value[key]
                    value = int(value)
                except Exception:
                    if failure_mode == FailureMode.THROW_EXCEPTION:
                        raise
                    elif failure_mode == FailureMode.LOG_ERROR:
                        log.error(
                            f"Error extracting time from JSON statistics, "
                            f"keys: {keys}"
                        )
                    value = 0
                return value

            # If the batch ended due to a delete operation, we have two
            # operations (and two statistics), otherwise only one.
            for i, stats in enumerate(result):
                # Show statistics of the update operation.
                try:
                    ins_after = stats["delta-triples"]["after"]["inserted"]
                    del_after = stats["delta-triples"]["after"]["deleted"]
                    ops_after = stats["delta-triples"]["after"]["total"]
                    num_ins = int(
                        stats["delta-triples"]["operation"]["inserted"]
                    )
                    num_del = int(
                        stats["delta-triples"]["operation"]["deleted"]
                    )
                    num_ops = int(stats["delta-triples"]["operation"]["total"])
                    time_ms = get_time_ms(stats, "total")
                    time_us_per_op = (
                        int(1000 * time_ms / num_ops) if num_ops > 0 else 0
                    )
                    if args.verbose == "yes":
                        log.info(
                            colored(
                                f"TRIPLES: {num_ops:+10,} -> {ops_after:10,}, "
                                f"INS: {num_ins:+10,} -> {ins_after:10,}, "
                                f"DEL: {num_del:+10,} -> {del_after:10,}, "
                                f"TIME: {time_ms:7,}ms, "
                                f"TIME/TRIPLE: {time_us_per_op:6,}µs",
                                attrs=["bold"],
                            )
                        )

                    # Also show a detailed breakdown of the total time.
                    time_parsing = (
                        get_time_ms(
                            stats,
                            "parsing",
                            failure_mode=FailureMode.SILENTLY_RETURN_ZERO,
                        )
                        if i == 0
                        else 0
                    )
                    time_planning = get_time_ms(stats, "planning")
                    try:
                        time_preparation = get_time_ms(
                            stats,
                            "execution",
                            "processUpdateImpl",
                            "preparation",
                            "total",
                            failure_mode=FailureMode.THROW_EXCEPTION,
                        )
                    except Exception:
                        time_preparation = get_time_ms(
                            stats,
                            "execution",
                            "processUpdateImpl",
                            "preparation",
                        )
                    time_where = get_time_ms(
                        stats,
                        "execution",
                        "processUpdateImpl",
                        "materializeResult",
                    )
                    time_metadata = get_time_ms(
                        stats,
                        "execution",
                        "updateMetadata",
                        failure_mode=FailureMode.SILENTLY_RETURN_ZERO,
                    )
                    time_insert = get_time_ms(
                        stats,
                        "execution",
                        "processUpdateImpl",
                        "insertTriples",
                        "total",
                        failure_mode=FailureMode.SILENTLY_RETURN_ZERO,
                    )
                    time_delete = get_time_ms(
                        stats,
                        "execution",
                        "processUpdateImpl",
                        "deleteTriples",
                        "total",
                        failure_mode=FailureMode.SILENTLY_RETURN_ZERO,
                    )
                    time_snapshot = get_time_ms(
                        stats, "execution", "snapshotCreation"
                    )
                    time_writeback = get_time_ms(
                        stats, "execution", "diskWriteback"
                    )
                    time_unaccounted = time_ms - (
                        time_parsing
                        + time_planning
                        + time_preparation
                        + time_where
                        + time_metadata
                        + time_delete
                        + time_insert
                        + time_snapshot
                        + time_writeback
                    )
                    if args.verbose == "yes":
                        log.info(
                            f"PARSING: {100 * time_parsing / time_ms:2.0f}%, "
                            f"PLANNING: {100 * time_planning / time_ms:2.0f}%, "
                            f"PREPARATION: {100 * time_preparation / time_ms:2.0f}%, "
                            f"WHERE: {100 * time_where / time_ms:2.0f}%, "
                            f"DELETE: {100 * time_delete / time_ms:2.0f}%, "
                            f"INSERT: {100 * time_insert / time_ms:2.0f}%, "
                            f"METADATA: {100 * time_metadata / time_ms:2.0f}%, "
                            f"SNAPSHOT: {100 * time_snapshot / time_ms:2.0f}%, "
                            f"WRITEBACK: {100 * time_writeback / time_ms:2.0f}%, "
                            f"UNACCOUNTED: {100 * time_unaccounted / time_ms:2.0f}%",
                        )

                    # Update the totals.
                    total_num_ops += num_ops
                    total_time_s += time_ms / 1000.0
                    elapsed_time_s = time.perf_counter() - start_time
                    time_us_per_op = int(1e6 * total_time_s / total_num_ops)

                except Exception as e:
                    log.warn(
                        f"Error extracting statistics: {e}, "
                        f"curl command was: {curl_cmd}"
                    )
                    # Show traceback for debugging.
                    import traceback

                    traceback.print_exc()
                    log.info("")
                    continue

            # Show statistics for the completed batch.
            if args.verbose == "yes":
                log.info(
                    colored(
                        f"TOTAL TRIPLES SO FAR: {total_num_ops:10,}, "
                        f"TOTAL UPDATE TIME SO FAR: {total_time_s:4.0f}s, "
                        f"ELAPSED TIME SO FAR: {elapsed_time_s:4.0f}s, "
                        f"AVG TIME/TRIPLE SO FAR: {time_us_per_op:,}µs",
                        attrs=["bold"],
                    )
                )
                log.info("")

            # Close the source connection (for each batch, we open a new one,
            # either from `event_id_for_next_batch` or from `since`).
            source.close()

            # After the first batch is processed, enable offset checking for
            # subsequent batches.
            first_batch = False

            # If Ctrl+C was pressed, we reached `--until`, or we processed
            # exactly `--num-messages`, finish.
            if (
                self.ctrl_c_pressed
                or self.finished
                or (
                    args.num_messages is not None
                    and total_num_messages >= args.num_messages
                )
            ):
                break

        # Final statistics after all batches have been processed.
        elapsed_time_s = time.perf_counter() - start_time
        time_us_per_op = (
            int(1e6 * total_time_s / total_num_ops) if total_num_ops > 0 else 0
        )
        log.info(
            f"Processed {batch_count} "
            f"{'batches' if batch_count > 1 else 'batch'} "
            f"terminating update command"
        )
        log.info(
            colored(
                f"TOTAL TRIPLES: {total_num_ops:10,}, "
                f"TOTAL TIME: {total_time_s:4.0f}s, "
                f"ELAPSED TIME: {elapsed_time_s:4.0f}s, "
                f"AVG TIME/TRIPLE: {time_us_per_op:,}µs",
                attrs=["bold"],
            )
        )
        return True



from __future__ import annotations

from qlever.command import QleverCommand
from qlever.commands.status import StatusCommand
from qlever.containerize import Containerize
from qlever.log import log
from qlever.util import stop_process_with_regex


def stop_container(server_container: str) -> bool:
    """
    Try to stop and remove container. return True iff it was stopped
    successfully. Gives log info accordingly.
    """
    for container_system in Containerize.supported_systems():
        if Containerize.stop_and_remove_container(
            container_system, server_container
        ):
            log.info(
                f"{container_system.capitalize()} container with "
                f'name "{server_container}" stopped and removed'
            )
            return True
    return False


class StopCommand(QleverCommand):
    """
    Class for executing the `stop` command.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return "Stop QLever server for a given dataset or port"

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str: list[str]]:
        return {"data": ["name"],
                "server": ["port"],
                "runtime": ["server_container"]}

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument("--cmdline-regex",
                               default="ServerMain.* -i [^ ]*%%NAME%%",
                               help="Show only processes where the command "
                                    "line matches this regex")
        subparser.add_argument("--no-containers", action="store_true",
                               default=False,
                               help="Do not look for containers, only for "
                                    "native processes")

    def execute(self, args) -> bool:
        # Show action description.
        cmdline_regex = args.cmdline_regex.replace("%%NAME%%", args.name)
        description = f"Checking for processes matching \"{cmdline_regex}\""
        if not args.no_containers:
            description += (f" and for Docker container with name "
                            f"\"{args.server_container}\"")
        self.show(description, only_show=args.show)
        if args.show:
            return True

        # First check if there is container running and if yes, stop and remove
        # it (unless the user has specified `--no-containers`).
        if not args.no_containers:
            if stop_container(args.server_container):
                return True

        # Check if there is a process running on the server port using psutil.
        # NOTE: On MacOS, some of the proc's returned by psutil.process_iter()
        # no longer exist when we try to access them, so we just skip them.
        stop_process_results = stop_process_with_regex(cmdline_regex)
        if stop_process_results is None:
            return False
        if len(stop_process_results) > 0:
            return all(stop_process_results)

        # If no matching process found, show a message and the output of the
        # status command.
        message = "No matching process found" if args.no_containers else \
            "No matching process or container found"
        log.error(message)
        args.cmdline_regex = "^ServerMain.* -i [^ ]*"
        log.info("")
        StatusCommand().execute(args)
        return True



from __future__ import annotations

import json

from termcolor import colored

from qlever.command import QleverCommand
from qlever.log import log
from qlever.qleverfile import Qleverfile
from qlever.util import run_command


class SettingsCommand(QleverCommand):
    """
    Class for executing the `settings` command.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return "Show or set server settings (after `qlever start`)"

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str, list[str]]:
        return {"server": ["port", "host_name", "access_token"]}

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "runtime_parameters",
            nargs="*",
            help="Space-separated list of runtime parameters to set "
            "in the form `key=value`; afterwards shows all settings, "
            "with the changed ones highlighted",
        ).completer = lambda **kwargs: [
            f"{key}=" for key in Qleverfile.SERVER_RUNTIME_PARAMETERS
        ]
        subparser.add_argument(
            "--endpoint_url",
            type=str,
            help="An arbitrary endpoint URL "
            "(overriding the one in the Qleverfile)",
        )

    def execute(self, args) -> bool:
        # Get endpoint URL from command line or Qleverfile.
        if args.endpoint_url:
            endpoint_url = args.endpoint_url
        else:
            endpoint_url = f"http://{args.host_name}:{args.port}"

        # Construct the `curl` commands for setting and getting.
        curl_cmds_setting = []
        keys_set = set()
        if args.runtime_parameters:
            for key_value_pair in args.runtime_parameters:
                try:
                    key, value = key_value_pair.split("=")
                except ValueError:
                    log.error("Runtime parameter must be given as `key=value`")
                    return False
                curl_cmds_setting.append(
                    f"curl -s {endpoint_url} -w %{{http_code}}"
                    f' --data-urlencode "{key}={value}"'
                    f' --data-urlencode "access-token={args.access_token}"'
                )
                keys_set.add(key)
        curl_cmd_getting = (
            f"curl -s {endpoint_url} -w %{{http_code}}"
            f" --data-urlencode cmd=get-settings"
        )
        self.show(
            "\n".join(curl_cmds_setting + [curl_cmd_getting]),
            only_show=args.show,
        )
        if args.show:
            return True

        # Execute the `curl` commands for setting the key-value pairs if any.
        for curl_cmd in curl_cmds_setting:
            try:
                curl_result = run_command(curl_cmd, return_output=True)
                body, http_code = curl_result[:-3], curl_result[-3:]
                if http_code != "200":
                    raise Exception(body)
            except Exception as e:
                log.error(
                    f"curl command for setting key-value pair failed: {e}"
                )
                return False

        # Execute the `curl` commands for getting the settings.
        try:
            curl_result = run_command(curl_cmd_getting, return_output=True)
            body, http_code = curl_result[:-3], curl_result[-3:]
            if http_code != "200":
                raise Exception(body)
            settings_dict = json.loads(body)
            if isinstance(settings_dict, list):
                settings_dict = settings_dict[0]
        except Exception as e:
            log.error(f"curl command for getting settings failed: {e}")
            return False
        for key, value in settings_dict.items():
            print(
                colored(
                    f"{key:<45}: {value}",
                    "blue" if key in keys_set else None,
                )
            )

        # That's it.
        return True



from __future__ import annotations

import sys
from importlib import import_module
from pathlib import Path


# Helper function to turn "snake_case" into "CamelCase".
def snake_to_camel(str):
    # Split by _ and - and capitalize each word.
    return "".join([w.capitalize() for w in str.replace("-", "_").split("_")])


# Get the name of the script (without the path and without the extension).
script_name = Path(sys.argv[0]).stem

ENGINE_NAMES = {
    "qlever": "QLever",
    "qmdb": "MillenniumDB",
}
# Default engine_name = script_name without starting 'q' and capitalized
engine_name = ENGINE_NAMES.get(script_name, script_name[1:].capitalize())

# Each module in `qlever/commands` corresponds to a command. The name
# of the command is the base name of the module file.
package_path = Path(__file__).parent.parent / script_name
command_names = [
    Path(p).stem
    for p in package_path.glob("commands/*.py")
    if p.name != "__init__.py"
]

# Dynamically load all the command classes and create an object for each.
command_objects = {}
for command_name in command_names:
    module_path = f"{script_name}.commands.{command_name}"
    try:
        module = import_module(module_path)
    except ImportError as e:
        raise Exception(
            f"Could not import module {module_path} for {script_name}: {e}"
        ) from e
    # Create an object of the class and store it in the dictionary. For the
    # commands, take - instead of _.
    class_name = snake_to_camel(command_name) + "Command"
    command_class = getattr(module, class_name)
    command_objects[command_name.replace("_", "-")] = command_class()



from __future__ import annotations

import psutil

from qlever.command import QleverCommand
from qlever.util import show_process_info


class StatusCommand(QleverCommand):
    """
    Class for executing the `status` command.
    """

    def __init__(self):
        pass

    def description(self) -> str:
        return ("Show QLever processes running on this machine")

    def should_have_qleverfile(self) -> bool:
        return False

    def relevant_qleverfile_arguments(self) -> dict[str: list[str]]:
        return {}

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument("--cmdline-regex",
                               default="^(ServerMain|IndexBuilderMain)",
                               help="Show only processes where the command "
                                    "line matches this regex")

    def execute(self, args) -> bool:
        # Show action description.
        self.show(f"Show all processes on this machine where "
                  f"the command line matches {args.cmdline_regex}"
                  f" using Python's psutil library", only_show=args.show)
        if args.show:
            return True

        # Show the results as a table.
        num_processes_found = 0
        for proc in psutil.process_iter():
            show_heading = num_processes_found == 0
            process_shown = show_process_info(proc, args.cmdline_regex,
                                              show_heading=show_heading)
            if process_shown:
                num_processes_found += 1
        if num_processes_found == 0:
            print("No processes found")
        return True



from __future__ import annotations

import json
import re
import shlex
import time

from qlever.command import QleverCommand
from qlever.log import log
from qlever.util import (
    run_command,
)


class MaterializedViewCommand(QleverCommand):
    """
    Class for executing the `materialized-view` command.
    """

    def __init__(self):
        self.materialized_view_name_regex = r"^[A-Za-z0-9-]+$"
        pass

    def description(self) -> str:
        return "Create a materialized view from the given query"

    def should_have_qleverfile(self) -> bool:
        return True

    def relevant_qleverfile_arguments(self) -> dict[str : list[str]]:
        return {
            "data": ["name"],
            "server": ["host_name", "port", "access_token"],
        }

    def additional_arguments(self, subparser) -> None:
        subparser.add_argument(
            "view_name",
            type=str,
            help="Name of the materialized view",
        )
        subparser.add_argument(
            "view_query",
            type=str,
            help="SPARQL query from which to create the materialized view",
        )
        subparser.add_argument(
            "--sparql-endpoint",
            type=str,
            help="URL of the SPARQL endpoint (default: <host_name>:<port>)",
        )

    def execute(self, args) -> bool:
        # SPARQL endpoint to use.
        sparql_endpoint = (
            args.sparql_endpoint
            if args.sparql_endpoint is not None
            else f"{args.host_name}:{args.port}"
        )

        # Check that the name of the materialized view is valid.
        if not re.match(self.materialized_view_name_regex, args.name):
            log.error(
                f"The name for the materialized view must match "
                f"the regex {self.materialized_view_name_regex}"
            )
            return False

        # Command for building the materialized view.
        url = (
            f"{sparql_endpoint}"
            f"?cmd=write-materialized-view"
            f"&view-name={args.view_name}"
        )
        materialized_view_cmd = (
            f"curl -s {shlex.quote(url)} "
            f"-H 'Authorization: Bearer {args.access_token}' "
            f"-H 'Content-type: application/sparql-query' "
            f"-d {shlex.quote(args.view_query)}"
        )
        self.show(materialized_view_cmd, args.show)
        if args.show:
            return True

        # Run the command (and time it).
        time_start = time.monotonic()
        try:
            log.info("Creating the materialized view ... "
                     "(this may take a while, depending on the complexity "
                     "of the query and the size of the result)")
            log.info("")
            result = run_command(materialized_view_cmd, return_output=True)
        except Exception as e:
            log.error(f"Creating the materialized view failed: {e}")
            return False
        time_end = time.monotonic()
        duration_seconds = round(time_end - time_start)

        # Try to parse the result (should be JSON).
        try:
            result_json = json.loads(result)
            view_name = result_json.get("materialized-view-written")
            log.info(
                f"Materialized view '{view_name}' created successfully "
                f"in {duration_seconds:,} seconds"
            )
        except Exception as e:
            log.error(f'Failed to parse JSON from "{result}": {e}')

        return True
